package org.zkoss.chart.options3D;

import org.zkoss.chart.Optionable;
import org.zkoss.chart.PlotAttribute;
import org.zkoss.chart.util.DynamicalAttribute;

/**
 * A 3D options.
 * <p>
 * All the options in this class support {@link DynamicalAttribute}.
 * 
 * @author hanhsu
 */
public class Options3D extends Optionable {
	private enum Attrs implements PlotAttribute, DynamicalAttribute {
		alpha, axisLabelPosition, beta, depth, enabled, fitToPlot, frame, viewDistance
	}
	
	/**
	 * Returns the value of alpha, one of the two rotation angles for the chart.
	 * <p>
	 * Default: 0.
	 */
	public Number getAlpha() {
		return getAttr(Attrs.alpha, 0).asNumber();
	}
	
	/**
	 * Sets the value of alpha
	 * 
	 * @param alpha
	 *            One of the two rotation angles for the chart. Defaults to 0.
	 */
	public void setAlpha(Number alpha) {
		setAttr(Attrs.alpha, alpha, 0);
	}

	/**
	 * Set it to "auto" to automatically move the labels to the best edge.
	 * <p>
	 * Default: null
	 * @since 7.2.1.0
	 */
	public String getAxisLabelPosition() {
		return getAttr(Attrs.axisLabelPosition, null).asString();
	}

	/**
	 * Set it to "auto" to automatically move the labels to the best edge.
	 * @since 7.2.1.0
	 */
	public void setAxisLabelPosition(String axisLabelPosition) {
		setAttr(Attrs.axisLabelPosition, axisLabelPosition);
	}

	/**
	 * Returns the value of beta, one of the two rotation angles for the chart.
	 * <p>
	 * Default: 0.
	 */
	public Number getBeta() {
		return getAttr(Attrs.beta, 0).asNumber();
	}

	/**
	 * Sets the value of beta
	 * 
	 * @param beta
	 *            One of the two rotation angles for the chart. Defaults to 0.
	 */
	public void setBeta(Number beta) {
		setAttr(Attrs.beta, beta, 0);
	}
	
	/**
	 * Returns the total depth of the chart.
	 * <p>
	 * Default: 100.
	 */
	public Number getDepth() {
		return getAttr(Attrs.depth, 100).asNumber();
	}

	/**
	 * Sets the total depth of the chart.
	 * 
	 * @param depth
	 *            the total depth of the chart. Defaults to 100.
	 */
	public void setDepth(Number depth) {
		setAttr(Attrs.depth, depth, 100);
	}
	
	/**
	 * Returns whether to render the chart using the 3D functionality.
	 * <p>
	 * Default: false.
	 */
	public boolean isEnabled() {
		return getAttr(Attrs.enabled, false).asBoolean();
	}

	/**
	 * Sets whether to render the chart using the 3D functionality.
	 * 
	 * @param enabled
	 *            whether to render the chart using the 3D functionality.
	 */
	public void setEnabled(boolean enabled) {
		setAttr(Attrs.enabled, enabled);
	}

	/**
	 * Returns whether the 3d box should automatically adjust to the chart plot area.
	 * <p>
	 * Default: true
	 * @since 7.2.1.0
	 */
	public boolean isFitToPlot() {
		return getAttr(Attrs.fitToPlot, true).asBoolean();
	}

	/**
	 * Sets whether the 3d box should automatically adjust to the chart plot area.
	 * @since 7.2.1.0
	 */
	public void setFitToPlot(boolean fitToPlot) {
		setAttr(Attrs.fitToPlot, fitToPlot);
	}
	
	/**
	 * Returns the frame options of the 3D chart.
	 * <p>
	 * Default: an instance of {@link Frame3D}.
	 */
	public Frame3D getFrame() {
		Frame3D frame = (Frame3D) getAttr(Attrs.frame);
		if (frame == null) {
			frame = new Frame3D();
			setFrame(frame);
		}
		return frame;
	}
	
	/**
	 * Sets the frame options of the 3D chart.
	 * @param frame
	 *            The options of drawing a frame around the 3D chart.
	 */
	public void setFrame(Frame3D frame) {
		setAttr(Attrs.frame, frame);
	}
	
	/**
	 * Returns the distance the viewer is standing in front of the chart.
	 * It is not used for 3D pie charts.
	 * <p>
	 * Default: 100.
	 */
	public Number getViewDistance() {
		return getAttr(Attrs.viewDistance, null).asNumber();
	}

	/**
	 * Sets the distance the viewer is standing in front of the chart.
	 * It is not used for 3D pie charts.
	 * 
	 * @param viewDistance
	 *            The distance the viewer is standing in front of the chart.
	 *            It is not usd for 3D pie charts
	 */
	public void setViewDistance(Number viewDistance) {
		setAttr(Attrs.viewDistance, viewDistance, 100);
	}
}
