/* PackedBubbleParentNodeOptions.java

	Purpose:
		
	Description:
		
	History:
		Wed Nov 20 15:52:22 CST 2019, Created by rudyhuang

Copyright (C) 2019 Potix Corporation. All Rights Reserved.
*/
package org.zkoss.chart.plotOptions;

import org.zkoss.chart.Marker;
import org.zkoss.chart.Optionable;
import org.zkoss.chart.PlotAttribute;
import org.zkoss.chart.util.DynamicalAttribute;

/**
 * Layout algorithm options for parent nodes.
 *
 * @author rudyhuang
 * @since 7.2.1.0
 */
public class PackedBubbleParentNodeOptions extends Optionable {
	private enum Attrs implements PlotAttribute, DynamicalAttribute {
		bubblePadding, friction, gravitationalConstant, initialPositionRadius,
		integration, linkLength, marker, maxIterations, maxSpeed,
		seriesInteraction, type
	}

	/**
	 * Returns the distance between two bubbles, when the algorithm starts to
	 * treat two bubbles as overlapping. The bubblePadding is also the expected
	 * distance between all the bubbles on simulation end.
	 * <p>
	 * Default: 5
	 */
	public Number getBubblePadding() {
		return getAttr(Attrs.bubblePadding, 5).asNumber();
	}

	/**
	 * Sets the distance between two bubbles, when the algorithm starts to
	 * treat two bubbles as overlapping. The bubblePadding is also the expected
	 * distance between all the bubbles on simulation end.
	 */
	public void setBubblePadding(Number bubblePadding) {
		setAttr(Attrs.bubblePadding, bubblePadding);
	}

	/**
	 * Returns the friction applied on forces to prevent nodes rushing to fast to the desired positions.
	 * <p>
	 * Default: -0.981.
	 */
	public Number getFriction() {
		return getAttr(Attrs.friction, -0.981).asNumber();
	}

	/**
	 * Sets the friction applied on forces to prevent nodes rushing to fast to the desired positions.
	 */
	public void setFriction(Number friction) {
		setAttr(Attrs.friction, friction);
	}

	/**
	 * Returns the gravitational const used in the barycenter force of the algorithm.
	 * <p>
	 * Default: 0.01
	 */
	public Number getGravitationalConstant() {
		return getAttr(Attrs.gravitationalConstant, 0.01).asNumber();
	}

	/**
	 * Sets the gravitational const used in the barycenter force of the algorithm.
	 */
	public void setGravitationalConstant(Number gravitationalConstant) {
		setAttr(Attrs.gravitationalConstant, gravitationalConstant);
	}

	/**
	 * When initialPositions are set to 'circle',
	 * initialPositionRadius is a distance from the center of circle, in which nodes are created.
	 * <p>
	 * Default: 20
	 */
	public Number getInitialPositionRadius() {
		return getAttr(Attrs.initialPositionRadius, 20).asNumber();
	}

	/**
	 * When initialPositions are set to 'circle',
	 * initialPositionRadius is a distance from the center of circle, in which nodes are created.
	 */
	public void setInitialPositionRadius(Number initialPositionRadius) {
		setAttr(Attrs.initialPositionRadius, initialPositionRadius);
	}

	/**
	 * Returns the integration type.
	 * Available options are 'euler' and 'verlet'.
	 * Integration determines how forces are applied on particles.
	 * In Euler integration, force is applied direct as {@code newPosition += velocity;}.
	 * In Verlet integration, new position is based on a previous position without
	 * velocity: {@code newPosition += previousPosition - newPosition}.
	 *
	 * Note that different integrations give different results as forces are different.
	 * <p>
	 * Default: "euler"
	 */
	public String getIntegration() {
		return getAttr(Attrs.integration, "euler").asString();
	}

	/**
	 * Sets the integration type.
	 * Available options are 'euler' and 'verlet'.
	 * Integration determines how forces are applied on particles.
	 * In Euler integration, force is applied direct as {@code newPosition += velocity;}.
	 * In Verlet integration, new position is based on a previous position without
	 * velocity: {@code newPosition += previousPosition - newPosition}.
	 *
	 * Note that different integrations give different results as forces are different.
	 */
	public void setIntegration(String integration) {
		if (!("euler".equals(integration) || "verlet".equals(integration)))
			throw new IllegalArgumentException("Unsupported integration: [ " + integration + " ]");
		setAttr(Attrs.integration, integration);
	}

	/**
	 * Returns the ideal length (px) of the link between two nodes.
	 * When not defined, length is calculated as: {@code Math.pow(availableWidth * availableHeight / nodesLength, 0.4);}
	 *
	 * Note: Because of the algorithm specification, length of each link might be not exactly as specified.
	 * <p>
	 * Default: null
	 */
	public Number getLinkLength() {
		return getAttr(Attrs.linkLength, null).asNumber();
	}

	/**
	 * Sets the ideal length (px) of the link between two nodes.
	 * When not defined, length is calculated as: {@code Math.pow(availableWidth * availableHeight / nodesLength, 0.4);}
	 *
	 * Note: Because of the algorithm specification, length of each link might be not exactly as specified.
	 */
	public void setLinkLength(Number linkLength) {
		setAttr(Attrs.linkLength, linkLength);
	}

	/**
	 * Returns the styling options for parentNodes markers, if any.
	 * Otherwise, create a new one.
	 * Similar to {@code line.marker} options.
	 */
	public Marker getMarker() {
		Marker marker = (Marker) getAttr(Attrs.marker);
		if (marker == null) {
			marker = new Marker();
			setMarker(marker);
		}
		return marker;
	}

	/**
	 * Sets the styling options for parentNodes markers.
	 * Similar to {@code line.marker} options.
	 */
	public void setMarker(Marker marker) {
		setAttr(Attrs.marker, marker);
	}

	/**
	 * Returns the max number of iterations before algorithm will stop.
	 * In general, algorithm should find positions sooner, but when rendering
	 * huge number of nodes, it is recommended to increase this value as finding
	 * perfect graph positions can require more time.
	 * <p>
	 * Default: 1000
	 */
	public Number getMaxIterations() {
		return getAttr(Attrs.maxIterations, 1000).asNumber();
	}

	/**
	 * Sets the max number of iterations before algorithm will stop.
	 * In general, algorithm should find positions sooner, but when rendering
	 * huge number of nodes, it is recommended to increase this value as finding
	 * perfect graph positions can require more time.
	 */
	public void setMaxIterations(Number maxIterations) {
		setAttr(Attrs.maxIterations, maxIterations);
	}

	/**
	 * Returns the max speed that node can get in one iteration.
	 * In terms of simulation, it's a maximum translation (in pixels) that a
	 * node can move (in both, x and y, dimensions).
	 * While friction is applied on all nodes, max speed is applied only for nodes
	 * that move very fast, for example small or disconnected ones.
	 * <p>
	 * Default: 5
	 */
	public Number getMaxSpeed() {
		return getAttr(Attrs.maxSpeed, 5).asNumber();
	}

	/**
	 * Sets the max speed that node can get in one iteration.
	 * In terms of simulation, it's a maximum translation (in pixels) that a
	 * node can move (in both, x and y, dimensions).
	 * While friction is applied on all nodes, max speed is applied only for nodes
	 * that move very fast, for example small or disconnected ones.
	 */
	public void setMaxSpeed(Number maxSpeed) {
		setAttr(Attrs.maxSpeed, maxSpeed);
	}

	/**
	 * Returns whether series should interact with each other or not.
	 * When parentNodeLimit is set to true, thi option should be set to false
	 * to avoid sticking points in wrong series parentNode.
	 * <p>
	 * Default: true
	 */
	public boolean isSeriesInteraction() {
		return getAttr(Attrs.seriesInteraction, false).asBoolean();
	}

	/**
	 * Sets whether series should interact with each other or not.
	 * When parentNodeLimit is set to true, thi option should be set to false
	 * to avoid sticking points in wrong series parentNode.
	 */
	public void setSeriesInteraction(boolean seriesInteraction) {
		setAttr(Attrs.seriesInteraction, seriesInteraction);
	}

	/**
	 * Returns the type of the algorithm used when positioning nodes.
	 * <p>
	 * Default: "reingold-fruchterman"
	 */
	public String getType() {
		return getAttr(Attrs.type, "reingold-fruchterman").asString();
	}

	/**
	 * Sets the type of the algorithm used when positioning nodes.
	 */
	public void setType(String type) {
		setAttr(Attrs.type, type);
	}
}
